package aQute.tester.testclasses.bundle.engine;

import org.junit.Test;

import junit.framework.TestCase;

// This test class is not supposed to be run directly; see readme.md for more info.
public class JUnit3And4Test extends TestCase {

	public void testSomething() {}

	@Test
	public void aJUnit4Test() {}

	@Test
	public void bJUnit4Test() {}
}
