package aQute.tester.testclasses.bundle.engine;

// This test class is not supposed to be run directly; see readme.md for more info.
public class JUnit5Test {
	@org.junit.jupiter.api.Test
	public void thisIsATest() {}

	@org.junit.jupiter.api.Test
	public void thisIsBTest() {}
}
