package aQute.tester.testclasses.bundle.engine;

import junit.framework.TestCase;

public class JUnit3AndVenusTest extends TestCase {

	@org.junit.venus.Test
	public void someTest() {}

	public void testSomethingElse() {

	}
}
