package aQute.tester.testclasses.bundle.engine;

import org.junit.Test;

// This test class is not supposed to be run directly; see readme.md for more info.
public class TestClass {
	@Test
	public void thisIsATest() {}
}
