package aQute.tester.test.assertions;

// This test class is not supposed to be run directly; see readme.md for more info.
public class CustomAssertionError extends AssertionError {
	private static final long serialVersionUID = 1L;
}